/* Name: main.c
 * Project: hid-data, example how to use HID for data transfer
 * Author: Christian Starkjohann
 * Creation Date: 2008-04-11
 * Tabsize: 4
 * Copyright: (c) 2008 by OBJECTIVE DEVELOPMENT Software GmbH
 * License: GNU GPL v2 (see License.txt), GNU GPL v3 or proprietary (CommercialLicense.txt)
 */

/*
This example should run on most AVRs with only little changes. No special
hardware resources except INT0 are used. You may have to change usbconfig.h for
different I/O pins for USB. Please note that USB D+ must be the INT0 pin, or
at least be connected to INT0 as well.
*/

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>  /* for sei() */
#include <util/delay.h>     /* for _delay_ms() */
#include <avr/eeprom.h>
#include <stdlib.h>

#include <avr/pgmspace.h>   /* required by usbdrv.h */
#include "usbdrv.h"
#include "i2csoft.h"

// ADDR R G B FL FH ACT
#define FLEN 0x07  /* Length of the feature report */

/* ------------------------------------------------------------------------- */
/* ----------------------------- USB interface ----------------------------- */
/* ------------------------------------------------------------------------- */

PROGMEM const char usbHidReportDescriptor[22] = {    /* USB report descriptor */
    0x06, 0x00, 0xff,              // USAGE_PAGE (Generic Desktop)
    0x09, 0x01,                    // USAGE (Vendor Usage 1)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x26, 0xff, 0x00,              //   LOGICAL_MAXIMUM (255)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x95, FLEN,                    //   REPORT_COUNT (8 bytes)
    0x09, 0x00,                    //   USAGE (Undefined)
    0xb2, 0x02, 0x01,              //   FEATURE (Data,Var,Abs,Buf)
    0xc0                           // END_COLLECTION
};
/* Since we define only one feature report, we don't use report-IDs (which
 * would be the first byte of the report). The entire report consists of 128
 * opaque data bytes.
 */

/* The following variables store the status of the current data transfer */
#define ACTION_NONE 0
#define ACTION_COLOR 0xFF
#define ACTION_RANDOM 0xEE
#define MAX_RAND 8
static uchar    currentAddress;
static uchar    bytesRemaining;
static uint8_t  freport[FLEN];
static uint8_t freportReceived = 0;
static uint8_t randomAddrs[MAX_RAND];
static uint8_t randomTimes[MAX_RAND];

#define COLORS 13
static uint32_t colors[COLORS] = { 0xFF0000, 0x00FF00, 0x0000FF,
		                   0xFFFF00, 0x00FFFF, 0xFF00FF,
                                   0xFFAA00, 0xAAFF00, 0xFF00AA,
                                   0xAA00FF, 0x00FFAA, 0x00AAFF,
                                   0xFFFFFF };

/* ------------------------------------------------------------------------- */

/* usbFunctionRead() is called when the host requests a chunk of data from
 * the device. For more information see the documentation in usbdrv/usbdrv.h.
 */
uchar   usbFunctionRead(uchar *data, uchar len)
{
    if(len > bytesRemaining)
        len = bytesRemaining;
    uint8_t read = 0;
    while (len > 0) {
      data[read++] = freport[currentAddress++];
      len--; bytesRemaining--;
    }
    return read;
}

/* usbFunctionWrite() is called when the host sends a chunk of data to the
 * device. For more information see the documentation in usbdrv/usbdrv.h.
 */
uchar   usbFunctionWrite(uchar *data, uchar len)
{
    if(bytesRemaining == 0)
        return 1;               /* end of transfer */
    if(len > bytesRemaining)
        len = bytesRemaining;
    uint8_t i;
    for (i = 0; i < len; i++) {
      freport[currentAddress + i] = data[i];
    }
    currentAddress += len;
    bytesRemaining -= len;
    if (bytesRemaining == 0) {
      freportReceived = 1;
      return 1; /* last chunk */
    }
    return 0;
}

/* ------------------------------------------------------------------------- */

usbMsgLen_t usbFunctionSetup(uchar data[8])
{
    usbRequest_t    *rq = (void *)data;

    if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS){    /* HID class request */
        if(rq->bRequest == USBRQ_HID_GET_REPORT){  /* wValue: ReportType (highbyte), ReportID (lowbyte) */
            /* since we have only one report type, we can ignore the report-ID */
            bytesRemaining = FLEN;
            currentAddress = 0;
            return USB_NO_MSG;  /* use usbFunctionRead() to obtain data */
        }else if(rq->bRequest == USBRQ_HID_SET_REPORT){
            /* since we have only one report type, we can ignore the report-ID */
            bytesRemaining = FLEN;
            currentAddress = 0;
            return USB_NO_MSG;  /* use usbFunctionWrite() to receive data from host */
        }
    }else{
        /* ignore vendor type requests, we don't use any */
    }
    return 0;
}

/* ------------------------------------------------------------------------- */

void setRGB(uint8_t addr, uint8_t r, uint8_t g, uint8_t b, uint8_t fadeL, uint8_t fadeH) {
  SoftI2CStart();
  SoftI2CWriteByte(addr << 1); // read, addr
  SoftI2CWriteByte(0x08); /* no idea anymore */
  SoftI2CWriteByte(r);
  SoftI2CWriteByte(g);
  SoftI2CWriteByte(b);
  SoftI2CWriteByte(fadeL);
  SoftI2CWriteByte(fadeH);
  SoftI2CWriteByte(0xFF); /* action */
  SoftI2CStop();
}

uint8_t getRandomIndex(uint8_t addr) {
  uint8_t i;
  for (i = 0; i < MAX_RAND; i++) {
    if (randomAddrs[i] == addr) return i;
  }
  return 0xFF;
}

uint8_t getFirstFreeRandomIndex() {
  uint8_t index = getRandomIndex(0xFF);
  if (index == 0xFF) return 0;
  return index;
}

void doRandom(uint8_t addr, uint8_t time) {
  uint32_t color = colors[random() % COLORS];
  setRGB(addr, (uint8_t)(color >> 16), (uint8_t)(color >> 8), (uint8_t)(color), 0x00, time);
}

uint16_t rtick = 0;
void randomTick() { // ~ called every 260ms
  uint8_t i;
  for (i = 0; i < MAX_RAND; i++) {
    uint8_t addr = randomAddrs[i];
    if (addr == 0xFF) continue;
    uint8_t time = randomTimes[i];
    if ((rtick % time) == 0) {
      doRandom(addr, time);
    }
  }
  rtick++;
}

int main(void)
{
    uchar   i;
    wdt_enable(WDTO_1S);
    wdt_reset();
    /* Even if you don't use the watchdog, turn it off here. On newer devices,
     * the status of the watchdog (on/off, period) is PRESERVED OVER RESET!
     */
    /* RESET status: all port bits are inputs without pull-up.
     * That's the way we need D+ and D-. Therefore we don't need any
     * additional hardware initialization.
     */
    for (i = 0; i < MAX_RAND; i++)
      randomAddrs[i] = 0xFF;      
    
    SoftI2CInit();
    usbInit();
    usbDeviceDisconnect();  /* enforce re-enumeration, do this while interrupts are disabled! */
    i = 0;
    while(--i){             /* fake USB disconnect for > 250 ms */
        wdt_reset();
        _delay_ms(1);
    }
    usbDeviceConnect();
    sei();
    uint8_t tick = 0;
    for(;;){                /* main event loop */
        wdt_reset();
        usbPoll();

        /* Any I2c Action? */
        if (freportReceived) {
          freportReceived = 0x00;
          uint8_t action = freport[6];
          if (action == ACTION_COLOR) {
            uint8_t ind = getRandomIndex(freport[0]);
            if (ind != 0xFF) {
              randomAddrs[ind] = 0xFF; 
            }
            setRGB(freport[0], freport[1], freport[2], freport[3], freport[4], freport[5]);      
            setRGB(freport[0], freport[1], freport[2], freport[3], freport[4], freport[5]);      
          }
          /* Any Random action? */ 
          else if (freport[6] == ACTION_RANDOM) {
            uint32_t seed = (uint32_t)freport[1] | (((uint32_t)freport[2]) << 8) | (((uint32_t)freport[3]) << 16) | (((uint32_t)freport[4]) << 24);
            srand(seed);  

            uint8_t addr = freport[0];
            uint8_t time = freport[5];
            uint8_t ind = getRandomIndex(addr);
            if (ind == 0xFF) ind = getFirstFreeRandomIndex();
            if (time == 0) time = 1;
            randomAddrs[ind] = addr;
            randomTimes[ind] = time;
            doRandom(addr, time);
          }
        }
        
        if ((++tick) == 0)
          randomTick();
        _delay_ms(1);
    }
    return 0;
}

/* ------------------------------------------------------------------------- */
